﻿using BI.Configuration;
using DingTalk.Api;
using DingTalk.Api.Request;
using DingTalk.Api.Response;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Configuration;
using Microsoft.IdentityModel.Tokens;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.IdentityModel.Tokens.Jwt;
using System.Security.Claims;
using System.Text;

namespace BI.Controllers
{
    [Route("api/auth")]
    //[Route("api/[controller]")]
    [ApiController]
    public class AuthController : ControllerBase
    {
        private readonly IConfiguration _configuration;
        public AuthController(IConfiguration configuration)
        {
            _configuration = configuration ?? throw new ArgumentNullException(nameof(configuration));
        }
        [HttpGet("{loginTmpCode}")]
        public IActionResult DingDingLogin(string loginTmpCode)
        {
            //判断参数是否为空
            if (string.IsNullOrEmpty(loginTmpCode))
            {
                return BadRequest("temp code error");
            }

            string appId = _configuration["DingAuth:appId"];//suitKey
            string appSecret = _configuration["DingAuth:appSecret"];
            // 获取access_token，注意正式代码要有异常流处理
            DefaultDingTalkClient client = new("https://oapi.dingtalk.com/gettoken");
            OapiGettokenRequest request = new()
            {
                Appkey = appId,
                Appsecret = appSecret,
            };
            request.SetHttpMethod("GET");
            OapiGettokenResponse response = client.Execute(request);
            if (response.IsError) return BadRequest();

            string access_token = response.AccessToken;


            // 通过临时授权码获取授权用户的个人信息
            DefaultDingTalkClient client2 = new("https://oapi.dingtalk.com/sns/getuserinfo_bycode");
            OapiSnsGetuserinfoBycodeRequest reqBycodeRequest = new()
            {
                // 通过扫描二维码，跳转指定的redirect_uri后，向url中追加的code临时授权码
                TmpAuthCode = loginTmpCode
            };
            OapiSnsGetuserinfoBycodeResponse bycodeResponse = client2.Execute(reqBycodeRequest, appId, appSecret);
            if (bycodeResponse.IsError) return BadRequest();
            // 根据unionid获取userid
            string unionid = bycodeResponse.UserInfo.Unionid;
            DefaultDingTalkClient clientDingTalkClient = new("https://oapi.dingtalk.com/topapi/user/getbyunionid");
            OapiUserGetbyunionidRequest reqGetbyunionidRequest = new()
            {
                Unionid = unionid
            };
            OapiUserGetbyunionidResponse oapiUserGetbyunionidResponse = clientDingTalkClient.Execute(reqGetbyunionidRequest, access_token);
            if (oapiUserGetbyunionidResponse.IsError) return BadRequest();
            string userid = oapiUserGetbyunionidResponse.Result.Userid;


            // 根据userId获取用户信息
            DefaultDingTalkClient clientDingTalkClient2 = new("https://oapi.dingtalk.com/topapi/v2/user/get");
            OapiV2UserGetRequest reqGetRequest = new()
            {
                Language = "zh_CN",
                Userid = userid
            };
            reqGetRequest.SetHttpMethod("POST");
            OapiV2UserGetResponse rspGetResponse = clientDingTalkClient2.Execute(reqGetRequest, access_token);
            if (rspGetResponse.IsError) return BadRequest();

            //创建JWT用户信息
            var userInfo = rspGetResponse.Result;
            JwtUserInfo jwtUser = new()
            {
                Mobile = userInfo.Mobile,
                Email = userInfo.Email,
                UserName = userInfo.Name,
                Id = unionid
            };
            //获取生成jwttoken
            string jwtToken = GenerateJwtToken(jwtUser);
            return Ok(new AuthResult
            {
                Success = true,
                Token = jwtToken
            });
        }
        private string GenerateJwtToken(JwtUserInfo user)
        {
            //  从 appsettings.json 中读取SecretKey
            var key = Encoding.ASCII.GetBytes(_configuration["JwtConfig:Secret"]);
            //  定义需要使用到的Claims
            var claims = new[]
            {
                new Claim("Id", user.Id),
                new Claim("mobile", user.Mobile),
                new Claim("username", user.UserName),
                new Claim(JwtRegisteredClaimNames.Email,user.Email),
                new Claim(ClaimTypes.Role, "Admin"),
                // Jti 用于刷新 token，我们将在下一篇中讲到
                new Claim(JwtRegisteredClaimNames.Jti, Guid.NewGuid().ToString())
            };
            // 生成Credentials
            var signingCredentials = new SigningCredentials(new SymmetricSecurityKey(key), SecurityAlgorithms.HmacSha256Signature);
            //  根据以上组件，生成token
            var token = new JwtSecurityToken(
                issuer: _configuration["JwtConfig:Issuer"],
                audience: _configuration["JwtConfig:Audience"],
                claims: claims,
                notBefore: DateTime.Now,
                expires: DateTime.Now.AddMinutes(double.Parse(_configuration["JwtConfig:Expires"])),
                signingCredentials: signingCredentials
                );
            var jwtToken = new JwtSecurityTokenHandler().WriteToken(token);
            return jwtToken;
        }
    }
}
